﻿using IronPython.Hosting;
using IronPython.Runtime;
using Microsoft.Scripting.Hosting;
using System.Collections.Generic;
using System.ComponentModel;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using Telerik.Windows.Controls;
using Telerik.Windows.Controls.GridView;

namespace EvalColumnDemo
{
    public class EvaluatedColumn : GridViewColumn
    {
        public string Expression
        {
            get
            {
                return (string)GetValue(ExpressionProperty);
            }
            set
            {
                SetValue(ExpressionProperty, value);
            }
        }

        public static readonly DependencyProperty ExpressionProperty =
        DependencyProperty.Register("Expression", typeof(string), typeof(EvaluatedColumn), new PropertyMetadata(string.Empty, new PropertyChangedCallback(PropertiesChanged)));

        public string PythonModules
        {
            get
            {
                return (string)GetValue(PythonModulesProperty);
            }
            set
            {
                SetValue(PythonModulesProperty, value);
            }
        }

        public static readonly DependencyProperty PythonModulesProperty =
DependencyProperty.Register("PythonModules", typeof(string), typeof(EvaluatedColumn), new PropertyMetadata(string.Empty, new PropertyChangedCallback(PropertiesChanged)));

        public string DataFormatString
        {
            get
            {
                return (string)GetValue(DataFormatStringProperty);
            }
            set
            {
                SetValue(DataFormatStringProperty, value);
            }
        }

        public static readonly DependencyProperty DataFormatStringProperty =
        DependencyProperty.Register("DataFormatString", typeof(string), typeof(EvaluatedColumn), new PropertyMetadata(string.Empty, new PropertyChangedCallback(PropertiesChanged)));

        private static void PropertiesChanged(object sender, DependencyPropertyChangedEventArgs e)
        {
            var column = sender as EvaluatedColumn;
            if (column.DataControl != null)
            {
                column.DataControl.Rebind();
            }
        }

        public override FrameworkElement CreateCellElement(GridViewCell cell, object dataItem)
        {
            var textBlock = new TextBlock();

            textBlock.TextAlignment = this.TextAlignment;
            textBlock.TextDecorations = this.TextDecorations;
            textBlock.TextWrapping = this.TextWrapping;
            textBlock.TextTrimming = this.TextTrimming;

            textBlock.Tag = new ExpressionValueWrapper(dataItem, this.Expression, this.PythonModules.Split(','));
            
            textBlock.SetBinding(TextBlock.TextProperty, new Binding("Value") { Source = textBlock.Tag, StringFormat = this.DataFormatString });

            return textBlock;
        }

        public class ExpressionValueWrapper : INotifyPropertyChanged
        {
            private static ScriptEngine scriptEngine = Python.CreateEngine();

            private string expression;
            private IEnumerable<string> modules;
            private object dataItem;
            private object value;

            public ExpressionValueWrapper(object dataItem, string expression, IEnumerable<string> modules)
            {
                this.dataItem = dataItem;
                if (this.dataItem is INotifyPropertyChanged)
                {
                    (this.dataItem as INotifyPropertyChanged).PropertyChanged += this.ExpressionValueWrapper_PropertyChanged;
                }

                this.expression = expression;
                this.modules = modules;

                this.Value = this.ExecuteExpression();
            }

            private void ExpressionValueWrapper_PropertyChanged(object sender, PropertyChangedEventArgs e)
            {
                if (this.expression.Contains(e.PropertyName))
                {
                    this.Value = this.ExecuteExpression(); 
                }
            }

            private object ExecuteExpression()
            {
                var source = scriptEngine.CreateScriptSourceFromString(this.expression, Microsoft.Scripting.SourceCodeKind.AutoDetect);

                try
                {
                    return source.Execute(this.GetScope());
                }
                catch (UnboundNameException e)
                {
                    System.Diagnostics.Debug.WriteLine(e.Message);
                }
                catch
                {
                }

                return null;
            }

            public object Value
            {
                get
                {
                    return this.value;
                }
                set
                {
                    if (value != this.value)
                    {
                        this.value = value;
                        this.OnPropertyChanged("Value");
                    }
                }
            }

            private ScriptScope GetScope()
            {
                var scope = PrepareScope(this.modules);
                
                foreach (var property in dataItem.GetType().GetProperties())
                {
                    scope.SetVariable(property.Name, property.GetValue(this.dataItem, null));
                }

                return scope;
            }

            private static ScriptScope PrepareScope(IEnumerable<string> modules)
            {
                var scope = scriptEngine.CreateScope();
                
                foreach (var module in modules)
                {
                    var mod = scriptEngine.ImportModule(module);
                    foreach (var variable in mod.GetVariableNames())
                    {
                        scope.SetVariable(variable, mod.GetVariable(variable));
                    }
                }

                return scope;
            }

            private void OnPropertyChanged(string propertyName)
            {
                if (this.PropertyChanged != null)
                {
                    this.PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
                }
            }
        
            public event PropertyChangedEventHandler PropertyChanged;
        }
    }
}
